package jssc_test;
import jssc.SerialPortList;
import jssc.SerialPort;
import jssc.SerialPortException;
import java.util.*;
import java.io.*;
import jssc.SerialPortEvent;
import jssc.SerialPortEventListener;

public class JSSC_test {
    
    //Serial Communication Variables
    public static SerialPort serialPort;
    static int mask;
    
    //Setup the flag variable to prevent too much gcode dripping
    public static String inputStr = "ok";
    
    //Setup menu variable
    public static String menu = "menu";

    /*/////////////////////////////////////////////////////////////////////////    
    //Calibration Constants - Edit these for any change in tape and reels
    //*************************************************************************

    Legend
        1 = 330 ohm
        2 = 400k ohm
        3 = 10k ohm
        4 = 75k ohm
        5 = 100k ohm
        6 = 51k ohm
        7 = 25k ohm
        8 = 5k ohm
        9 = ohm
        10 = ohm
        11 = ohm
        12 = ohm
        13 = ohm
        14 = ohm
        15 = ohm
        16 = ohm
        17 = ohm
    
    Reel Tape X and Y - are the absolute positions from origin where the nozzle
    will go to induce a tape retraction. In other words, from the standard view
    the machine will move to the tape, lower z axis, and move to the left 4 mm.
    Change these values as you swap out the feeder guides.
    
    Reel Posit X and Y - are the RELATIVE positions from Reel Tape X and Y after
    before a retraction. This means if tapex and tapey were (100,100) before 
    retraction, the retraction will occur and the nozzle will move to
    (100 - reelPositX, 100 - reelPositY). Change these values if alignment is 
    off or you swap new components.
    
    Reel Height - is the z height above origin which the nozzle will touch the
    component on the tape. Adjust this accordingly to allow a little contact
    with the component. Crashing will not occur if you built the machine as I
    did. The weight of the zaxis guide rod, nozzle, and a little bit of spring
    force will not hurt your components.
    
    Steps to Advance - is the number of times you need pull the tape in 4mm 
    increments to expose a full new components. Adjust this for anything bigger
    than a 0805 tape (ex. soic8 has components centered at every 8mm, so two
    4mm  retraction cycles need to occur for machine to expose another component
    for placing
  
    */  
    
    public static final int xyspeed = 8000;  //max table speed in mm/min
    public static final int zspeed =  200;   //max z axis speed in mm/min
    public static final int maxx =    240;   //max positive x position in mm from origin
    public static final int maxy =    240;   //max positive y position in mm from origin
    public static final int maxz =    120;   //max positive z position in mm from origin
    public static final boolean invertNozzle = true; //invert nozzle operation
    public static final double retractHeight = 3.5;     //standard retraction height in mm
    public static final int retractionSpeed = 500; //max retraction speed in mm/min
    public static final double retractionX = -4.0;
    public static final double boardHeight = 0.25;         //target board height in seconds
    public static final int suctionDwell = 2000; //suction dwell period in seconds
    public static final int releaseDwell = 2000; //release dwell period in seconds
    public static final String fileDirectory = "C:\\Users\\DavidD\\Desktop\\PROJECT\\project_packages\\pickandplace\\programmed_files";
    
    public static final int reelTotal = 17;     //   1       2       3       4       5       6       7       8       9      10      11      12      13      14      15      16      17
    public static final double reelTapeX[] =  { 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50, 239.50};
    public static final double reelTapeY[] =  { 232.00, 220.00, 208.00, 196.25, 184.25, 172.25, 160.50, 148.50, 136.50, 124.50, 112.75, 100.75,  88.75,  76.75,  64.75,  52.75,  40.75};
    public static final double reelPositX[] = {      2,      2,      2,      2,      2,      2,      2,      2,      2,      2,      2,      2,      2,      2,      2,      2,      2};
    public static final double reelPositY[] = {    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5};
    public static final double reelHeight[] = {    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5,    1.5};
    public static final int stepsToAdvance[] ={      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1};

    //*************************************************************************
    ///////////////////////////////////////////////////////////////////////////
    
    //Main loop
    public static void main(String[] args) throws FileNotFoundException, IOException, SerialPortException {
        
        //Intro header
        Scanner console = new Scanner(System.in);
        System.out.println("<--- WELCOME TO PICK and PLACE COMMUNICATOR --->\n");
        
        //print name of all available ports to communicate with
        printPorts();
        
        //select port to open
        selectPortToOpen(console);
        
        //Main menu loops until exit is received
        while(!menu.equals("exit")){
            
            //Get what user wants to do
            inputSelection(console);
            
            //calibration check to perform slow table travel, fast table travel, then reelcheck
            if(menu.equals("calib")){
                System.out.println("Ensure you are zeroed and the table is clear, then press enter");
                System.in.read();
                tableExtents("slow", serialPort);
                System.out.println("Slow run complete, press enter to perform fast speed check");
                System.in.read();
                tableExtents("fast", serialPort);
                System.out.println("Fast run complete");
                System.out.println("Table Extents calibraiton complete\n\n");
            }
            
            //Perform Pick and Place of Programmed File
            if(menu.equals("run")){
                runFile(console);
            }
            
            //Manual Input Commands
            if(menu.equals("manual")){
                manualInput(console);
            }
            
            //alignment of tape and reels
            if(menu.equals("alignTape")){
                System.out.println("Performing Reel Check");
                reelCheck(console);
                System.out.println("Reel Check Complete");
            }
            
            //exit the program
            if(menu.equals("exit")){
                break;
            }
        }
    }
    
    //Printout the available ports
    public static void printPorts(){
        String[] portNames = SerialPortList.getPortNames();
        System.out.println("AVAILABLE PORTS:");
        for(int i = 0; i < portNames.length; i++){
            System.out.println(i+1 + ".) "+ portNames[i]);
        }
        System.out.println();
    }
    
    //Ask for port to open, and open it, set parameters for communicating
    public static void selectPortToOpen(Scanner console) throws SerialPortException{
        
        //Which Port to  open?
        System.out.print("ENTER PORT TO OPEN: " );
        String arduinoPort = console.next();
        System.out.println("You have chosen \"" + arduinoPort + "\" to communicate with\n");
        serialPort = new SerialPort(arduinoPort);
        
        //would you like to open port?
        System.out.print("OPEN PORT? (yes or no): ");
        String inputMode = console.next();
        System.out.println();
        
        //Open Port, set parameters
        if(inputMode.equals("yes")){
            try {
                System.out.println("Port opened: " + serialPort.openPort());
                //Set port paramters: baud, data bit length, stop bit, parity, setRTS, set DTR 
                System.out.println("Parameters set: " + serialPort.setParams(115200, 8, 1, 0, false,true));
                int mask = SerialPort.MASK_RXCHAR + SerialPort.MASK_CTS + SerialPort.MASK_DSR;//Prepare mask
                serialPort.setEventsMask(mask);//Set mask
                serialPort.addEventListener(new SerialPortReader()); //add event listener
            }
            catch (SerialPortException ex){
                System.out.println(ex);
            }
        }
        
        //Wait for PNP to spool up
        System.out.println("Machine Startup Routine Running. Please Wait");
        try{
            Thread.sleep(7000);
        }
        catch(Exception ex){
            System.out.println(ex);
        }
        
        //machine setup gcode commands
        pnpWrite("G21");  //set units to millimeters
        pnpWrite("G90");  //start in absolute coordinate system
        pnpWrite("G92 X0 Y0 Z0"); //tell machine it is at 0,0,0
        System.out.println("PNP GCode Setup Commands Sent\nMachine Ready\n\n");
    }
    
    //Select type of input
    public static void inputSelection(Scanner console) throws FileNotFoundException, IOException{
        int menunumber = 0;
        while(menunumber > 5 || menunumber < 1){
            System.out.print("\nWHAT WOULD YOU LIKE TO DO?"
                           + "\n 1. Calibrate Speed"
                           + "\n 2. Run File"
                           + "\n 3. Manual Input"
                           + "\n 4. Tape and Reel Alignment"
                           + "\n 5. Exit"
                           + "\n Enter Selection: ");
            menunumber = console.nextInt();
            System.out.println();
        }
        if(menunumber == 1){
            menu = "calib";
        }
        else
        if(menunumber == 2){
            menu = "run";
        }
        else
        if(menunumber == 3){
            menu = "manual";
        }else
        if(menunumber == 4){
            menu = "alignTape";
        }else
        if(menunumber == 5){
            menu = "exit";
        }
    }
    
    //Manual input of GCode to pnp
    public static void manualInput(Scanner console) throws IOException, SerialPortException{
        inputStr = "ok";
        //Introduction of how to put in coordinates
        System.out.println("Type in GCode to send to Pick and Place");
        System.out.println("type 'no' to exit at end of command. Example below\n");
        //Input Manual GCode
        do{
            System.out.print("Input G Code: ");
            BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
            String GCode = br.readLine();
            System.out.println(GCode);
            System.out.print("Is this correct? tye 'no' if not:  ");
            String evalString = console.next();
            if(evalString.equals("yes")){
                break;
            }
            System.out.println("Writing to Port\nPNP busy\n");
            pnpWrite(GCode);
        }
        while(true);
    }
    
    //go to table extents
    public static void tableExtents(String speed, SerialPort serialPort) throws SerialPortException{
        
        //adjust moveSpeed variable appropriately
        int moveSpeed = 0;
        int repeat = 1;
        if(speed.equals("slow")){
            moveSpeed = 2000; //millimeters per second
            repeat = 1;
        }else
        if(speed.equals("fast")){
            moveSpeed = xyspeed;
            repeat = 3;
        }
        
        //perform travel moves, more if doing fast speed test
        for(int i =0; i < repeat; i++){
            //Initilize variable
            String GCode = "";
            
            //lift nozzle to standard height to prevent crash
            GCode = "G1 Z" + Double.toString(retractHeight) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);

            //move to top left
            GCode = "G1 Y" + Integer.toString(maxy) + " F" + Integer.toString(moveSpeed);
            pnpWrite(GCode);

            //move to top right
            GCode = "G1 X" + Integer.toString(maxx) + " F" + Integer.toString(moveSpeed);
            pnpWrite(GCode);

            //move to bottom right
            GCode = "G1 Y0 F" + Integer.toString(moveSpeed);
            pnpWrite(GCode);

            //move to bottom left
            GCode = "G1 X0 F" + Integer.toString(moveSpeed);
            pnpWrite(GCode);

            //drop nozzle to 0
            GCode = "G1 Z0 F" + Integer.toString(zspeed);
            pnpWrite(GCode);
        }
    }
    
    //check positions of reels
    public static void reelCheck(Scanner console) throws SerialPortException, IOException{
    
        //initialize variable
        String GCode = "";

        //lift nozzle to prevent crash
        pnpWrite("G1 Z" + Double.toString(retractHeight) + " F" + Integer.toString(zspeed));
        
        //turn off suction
        if(invertNozzle == true){
            pnpWrite("M106");
        }else{
            pnpWrite("M107");
        }
        
        //go to each starting retraction point
        for(int i =0; i< reelTotal; i++){
            //go to position
            GCode = "G1 X" + Double.toString(reelTapeX[i]) + " Y" + Double.toString(reelTapeY[i]) + " F" + Integer.toString(xyspeed);
            pnpWrite(GCode);
            
            //drop nozzle to component height a distance for retraction verification
            double drop = reelHeight[i] - 0.5;
            GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //ask to advance
            System.out.println("Press enter to continue");
            System.in.read();
            
            //raise nozzle
            GCode = "G1 Z" + Double.toString(retractHeight) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //center on current component
//            double positx = reelTapeX[i] - reelPositX[i];
//            double posity = reelTapeY[i] + reelPositY[i];
//            GCode = "G1 X" + Double.toString(positx) + " Y" + Double.toString(posity) + " F" + Integer.toString(xyspeed);
//            pnpWrite(GCode);
            
            //drop nozzle to component height to check alignment
//            GCode = "G1 Z" + Double.toString(reelHeight[i]) + " F" + Integer.toString(zspeed);
//            pnpWrite(GCode);
            
            //ask to advance
//            System.out.println("Press enter to continue");
//            System.in.read();
            
            //raise nozzle
//            GCode = "G1 Z" + Double.toString(retractHeight) + " F" + Integer.toString(zspeed);
//            pnpWrite(GCode);
        }
        
        //return to origin
        pnpWrite("G1 X0 Y0 F" + Integer.toString(xyspeed));
    }
    
    //Run File execution
    public static void runFile(Scanner console) throws FileNotFoundException, SerialPortException{
        
        //ensure at origin
        pnpWrite("G1 X0 Y0 Z0 F1000");
        
        //turn off nozzle suction
        if(invertNozzle == true){
            pnpWrite("M106");
        }else{
            pnpWrite("M107");
        }
        
        //search directory and list names of files
        File folder = new File(fileDirectory);
        File[] listOfFiles = folder.listFiles();

        System.out.println("\nContents of Directory\n");
        for(int i = 0; i < listOfFiles.length; i++) {
            if (listOfFiles[i].isFile()) {
                System.out.println("File:    " + listOfFiles[i].getName());
            } else if (listOfFiles[i].isDirectory()) {
                System.out.println("Folder:  " + listOfFiles[i].getName());
            }
        }
        
        //ask which file to run
        System.out.print("Which File to run? include extension: ");
        String fileName = console.next();
        String fullpath = fileDirectory + "\\" + fileName;
        File GCodeFile = new File(fullpath);
        
        //Check if file exists
        while(!GCodeFile.exists()){
            System.out.println("File does not exist!!! ");
            System.out.print("Enter file name with extension: ");
            fileName = console.next();
            GCodeFile = new File(fileName);
        }
        System.out.println("File " + fileName + " opened.");
        
        //Create a follow up file that records all compeleted operations
        PrintStream newFile = new PrintStream("currentProject.txt");
        System.out.println("currentProject.txt was created \n");
        
        //ask to commence execution
        System.out.print("Commence Execution? (yes or no) ");
        fileName = console.next();
        if(fileName.equals("no"))
            return;
        
        //Open Scanner on file
        Scanner gcodeScanner = new Scanner(GCodeFile);
        
        /*
        The file format reads in a string at a time. The first token must be a
        number that corresponds with a tape position number, IE the precoded 
        positions from reelTotal
        
        No error checking is performed, recommend you perform a dry run prior to
        doing anything with the file to ensure your file is coded correclty
        
        If you dont know, a token is a continuous string of characters so each
        token is divided by spaces. The coded format is as follows:
        
        First Token - component from reel 1 will be retracted and picked up
        Second Token - x coordinate of where component is centered to go
        Third Token - y coordinate of where component is centered to go
        Fourth Token - Rotation of object prior to putting on board
        
        Example: "1 X95.00 Y90.00 S0"
        
        without the quotes, cnc will go to first reel, retract component,
        pick it up from the reel, move to (95,90), rotate to 0 degrees, and 
        place the component
        */
        
        //execute line by line
        while(gcodeScanner.hasNextLine()){
            
            //Consume line, copy it to backup file in case of mid pnp failure
            String incomingString = gcodeScanner.nextLine();
            newFile.println(incomingString);
            
            //Open scanner on string
            Scanner incomingToken = new Scanner(incomingString);

            //4 tokens are expected, put them in variables for manipulation
            String component = incomingToken.next();
            String xposit = incomingToken.next();
            String yposit = incomingToken.next();
            String rotation = incomingToken.next();
            
            //lift nozzle to standard height
            String GCode = "G1 Z" + Double.toString(retractHeight) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //move to tape position for retraction
            int pickUp = Integer.parseInt(component) - 1; //adjust because of array starts at 0
            GCode = "G1 X" + Double.toString(reelTapeX[pickUp]) + " Y" + Double.toString(reelTapeY[pickUp]) + " F" + Integer.toString(xyspeed);
            pnpWrite(GCode);
            
            //drop nozzle for retraction
            GCode = "G1 Z0.25 F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //retract tape and move back a little bit
            pnpWrite("G91"); //temporarily switch to relative coordinates
            GCode = "G1 X" + Double.toString(retractionX) + " F" + Integer.toString(retractionSpeed);
            pnpWrite(GCode);
            GCode = "G1 X0.75" + " F" + Integer.toString(retractionSpeed);
            pnpWrite(GCode);
            pnpWrite("G90"); //return to absolute coordinates
            
            //lift nozzle
            GCode = "G1 Z" + Double.toString(retractHeight) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //move over component
            pnpWrite("G91"); //temporarily switch to relative coordinates
            GCode = "G1 X" + Double.toString(reelPositX[pickUp]) +  " Y" + Double.toString(reelPositY[pickUp])  + " F" + Integer.toString(xyspeed);
            pnpWrite(GCode);
            pnpWrite("G90"); //return to absolute coordinates
            
            //drop nozzle on component to pick up
            double drop = reelHeight[pickUp] - 1;
            GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //turn on suction to pick up component
            if(invertNozzle == true){
                pnpWrite("M107");
            }else{
                pnpWrite("M106");
            }
            
            //dwell to ensure suction
            GCode = "G4 P" + Integer.toString(suctionDwell);
            pnpWrite(GCode);
            
            //lift nozzle to standard retraction
            GCode = "G1 Z" + Double.toString(retractHeight) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //move to xposit and yposit
            GCode = "G1 " + xposit + " " + yposit + " F" + Integer.toString(xyspeed);
            pnpWrite(GCode);
            
            //rotate code
            //rotate code
            //rotate code
            //rotate code
            //rotate code
            //rotate code
            //rotate code
            
            //drop nozzle to board height + component height
            drop = boardHeight;
            GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //turn off suction
            if(invertNozzle == true){
                pnpWrite("M106");
            }else{
                pnpWrite("M107");
            }
            
            //dwell to ensure release of component
            GCode = "G4 P" + Integer.toString(releaseDwell);
            pnpWrite(GCode);
        }
        System.out.println("\nFile Execution Complete. Returning to Home");
        pnpWrite("G1 Z3.5 F100");   //raise z axis
        pnpWrite("G1 X0 Y0 F2000"); //go to zero on x and y axis
        pnpWrite("G1 Z0 F100");     //go to zero on z axis
        pnpWrite("M107");           //turn off valve
        pnpWrite("M18");            //disable steppers
    }
    
    //Sleep Until Computer receives an "okay" from the PNP
    public static void pnpWrite(String GCode) throws SerialPortException{
        System.out.println("Sent: " + GCode);  //uncomment to see what is being sent to machine
        GCode = GCode + "\n\r"; //add the required new line and carriage return
        inputStr = "busy";
        serialPort.writeString(GCode); //write to the port
        //Stay in loop until okay from machine is received
        while(true){
            try{
                Thread.sleep(100);
                if(inputStr.equals("ok")){
                    break;
                }
            }
            catch(Exception ex){
                System.out.println(ex);
            }

        }
    }
    
    //Serial Port Listener
    static class SerialPortReader implements SerialPortEventListener {
        public void serialEvent(SerialPortEvent event) {    
            if(event.isRXCHAR()){
                try{
                    Thread.sleep(200); //dleay a little bit to get everything into the buffer
                }
                catch(Exception ex){
                    System.out.println(ex);
                }
                //Copy the input into decision variable if expected input recieved
                try {
                    String tempVal = serialPort.readString();
//                    System.out.print(tempVal);  //UNCOMMENT TO SEE WHAT IT IS BEING received
                    if("ok\n".equals(tempVal)){
                        inputStr = "ok";
                    }
                }
                catch (SerialPortException ex) {
                    System.out.println(ex);
                }
            }
         }
     }
    
    
}